/*
**	main.c
**
*/
#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        "dec.h"

/*
**
*/
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

static int a_fd= -1;
static int vol= 255;

void
set_volume(volume)/* 0-100*/
	{
	int fd, ret, devices, vol_chan;
	if ((fd= open ("/dev/mixer", O_WRONLY | O_NDELAY, 0)) < 0){
		fprintf(stderr,"audio device /dev/mixer is busy or can't be opened\n");
		exit(1);
		}
	ret= ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devices);
	if (ret== -1) devices = 0;
	if (devices & (1 << SOUND_MIXER_PCM)) vol_chan= SOUND_MIXER_PCM;
  			else		      vol_chan= SOUND_MIXER_VOLUME;
	volume |= volume << 8;
	if( ioctl(fd, MIXER_WRITE(vol_chan), &volume) == -1)
		{ fprintf(stderr,"configuration of /dev/mixer failed\n"); exit (1); }
	close(fd);
	}

void
open_audio(noc,freq,volume) {
static int p_pr;
static int p_st;
static int p_sr;
  int fd, flags;

  if ((fd = open ("/dev/dsp", O_WRONLY | O_NDELAY, 0)) < 0){
      fprintf(stderr,"audio device /dev/dsp is busy or can't be opened\n");
      exit(1);
    }

  /* turn NDELAY mode off: */
  if ((flags = fcntl (fd, F_GETFL, 0)) < 0) {
    fprintf(stderr,"fcntl F_GETFL on /dev/audio failed\n");
    exit(1);
  }
  flags &= ~O_NDELAY;
  if (fcntl (fd, F_SETFL, flags) < 0) {
    fprintf(stderr,"fcntl F_SETFL on /dev/audio failed\n");
    exit(1);
  }
  a_fd= fd;

  /* configure the device: */
  p_pr= 16;
  p_st= noc-1;
  p_sr= freq;
  if(
      ioctl(a_fd, SNDCTL_DSP_SAMPLESIZE, &p_pr) == -1 ||
      ioctl(a_fd, SNDCTL_DSP_STEREO, &p_st) == -1 ||
      ioctl(a_fd, SNDCTL_DSP_SPEED, &p_sr) == -1
    ) {
	  fprintf(stderr,"configuration of /dev/dsp failed\n");
	  exit (1); }
  set_volume(volume);
}

void
awrite(char *buf,int len) {
int n;
while(len) {
	if ((n= write (a_fd, buf, len)) < 0) {
		if (errno != EINTR)
		{
			perror("write to /dev/dsp failed");
			exit(1);}
		n= 0;
		}
	len-= n; buf+= n;
	}
}

static FILE	*fp,*wav;
static unsigned long	frameNum;
static frame_params	fr_ps;
char ename[64];
char dname[64];

int getc1(void){ return getc(fp); }

/*
**	au_out_sample(): output 32 sample (64 word)
**	The type of Sample is int. The value is 31 bit (normally -2^30 -- 2^30-1) 
**	Outside of the range should be clipped, and shifted left.
*/
void
au_out_sample(Sample *pcm)
{
	short pc[64]; int i;

	for(i=0;i<64;i++) {
		int r= pcm[i];
		r+= (1<<14);	// rounding before clip (for 16 bit precision)
#ifdef __arm__
		r= QADD(r,r);	// *2 with clipping
#else
		if (r<-(1<<30)) r= -(1LL<<31);		// clipping
		else if (r>(1<<30)-1) r= (1LL<<31)-1;
		else r<<= 1;				// *2
#endif
	        pc[i]= r>>16;				// save it to 16 bit array
	}
	if(wav) fwrite(pc, 2, 64, wav);
	else awrite((char *)pc, 64*sizeof(short));
}


static Us s_freq[2][4]= {
{22050, 24000, 16000, 0},	// MPEG-2 LSF
{44100, 48000, 32000, 0}	// MPEG-1
};
static Us bitrate[2][3][15]= {
{{0,32,48,56,64,80,96,112,128,144,160,176,192,224,256}, // MPEG-2 LSF
 {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
 {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}},
{{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448}, // MPEG-1
 {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
 {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}}
};

void WriteHdr(fr_ps, s)
frame_params *fr_ps;
FILE *s;
{
char *mode_names[5] = { "stereo", "j-stereo", "dual-ch", "single-ch" , "multi-ch"};
char *layer_names[3] = { "I", "II", "III" };
char *version_names[2] = { "MPEG-2 LSF", "MPEG-1" };

layer *info = fr_ps->header;

   fprintf(s, "HDR: s=FFF, id=%X, l=%X, ep=%s, br=%X, sf=%X, pd=%X, ",
           info->version, info->lay, ((info->error_protection) ? "on" : "off"),
           info->bitrate_index, info->sampling_frequency, info->padding);
   fprintf(s, "pr=%X, m=%X, js=%X, c=%X, o=%X, e=%X\n",
           info->extension, info->mode, info->mode_ext,
           info->copyright, info->original, info->emphasis);
   fprintf(s, "alg.=%s, layer=%s, tot bitrate=%d, sfrq=%d\n",
           version_names[info->version], layer_names[info->lay-1],
	   bitrate[info->version][info->lay-1][info->bitrate_index],
           s_freq[info->version][info->sampling_frequency]);
   fprintf(s, "mode=%s, sblim=%d, jsbd=%d, ch=%d\n",
           mode_names[info->mode], fr_ps->sblimit, fr_ps->jsbound, fr_ps->stereo);
   fflush(s);
}

void
s_err(int en){
	if(en==3) fprintf(stderr,"layer III is not supported\n");
	if(en==4) fprintf(stderr,"layer IV (AAC) is not supported\n");
}

void
au_frame_info(int lay, int sf, int st)
{
       if(frameNum++==0) {
#ifdef AUDIO
		open_audio(st, sf, vol);
#else
		{ strcpy(dname, ename); strcat(dname, ".dec"); }
#endif
		if(*dname) if (!(wav= fopen(dname,"w+b"))) {
			fprintf(stderr,"Can't create \"%s\".\n", dname);
			exit(1);
		}
#ifdef VERB2
       		WriteHdr(&fr_ps, stdout);  /* printout layer/mode */
#endif
       }
}

static
void usage(char *p){
fprintf(stderr,"Usage: %s input [output]\n", p);
fprintf(stderr,"input: bitstream of encoded audio\n");
fprintf(stderr,"output: WAV file \n");
fprintf(stderr,"(dflt /dev/dsp for linux soundcard)\n");
fprintf(stderr,"-vN Volume (N = [0-100])\n");
fprintf(stderr,"(dflt inName+%s)\n", ".dec");
exit(1);
}

int
main(int argc, char **argv)
{
	char *p= *argv;
	argc--; argv++;
	while (argc) {
	if(!strncmp(*argv,"-v",2))vol= atoi(*argv+2);
	else if(**argv=='-'){ fprintf(stderr, "unknown option: %s\n", *argv); usage(p); }
	else break;
	argc--; argv++;
	}
    if(!argc) fp= stdin; // else, use stdin for input bitstream
    else {
	if(argc==2) strcpy(dname, argv[1]);
	if(argc>2) usage(p);
	strcpy(ename, argv[0]);
	if(!(fp=fopen(ename,"rb"))){ fprintf(stderr,"%s not found\n", ename); exit(1); }
	}

    dec_frame(&fr_ps);		// task decoder routine

    if(fp && fp!=stdin) fclose(fp);
    if(wav) fclose(wav);
    return 0;
}
